
import numpy

class PbWriter:
    """PbWriter writes collections of particles in .pb format.  Use as in:
	pb = PbWriter( outfname, [list_of_fieldnames] )
	followed by one or more calls to write particle data, any number of particles at a time
	pb.writepcles( xyz, fields, id )  # xyz is Nx3 numpy array, fields is Nxnfields, id is 1D integers (or None)
	pb.writepcles( morexyz, morefields, id )
	...
	pb.close()
    """

    PBHMagic = 0xffffff98;

    def __init__(self, outf, fieldnames):
	if isinstance(outf, file):
	    self.outfname = outf.name
	    self.outf = outf
	else:
	    self.outfname = outf
	    self.outf = open(outf, 'wb')

	self.fieldnames = fieldnames
	self.pcount = 0

	flen = sum(map(len, fieldnames)) + len(fieldnames)
	pad = ((flen + 3) & ~3) - flen
	dataoff = 3*4 + flen + pad

	head = numpy.array( [ self.PBHMagic, dataoff, len(fieldnames) ], dtype=numpy.int32 )
	head.tofile( self.outf )
	fieldstr = "\0".join(fieldnames) + "\0" * (pad+1)
	self.outf.write( fieldstr )

	# now ready to append particles

    def writepcles(self, xyz, fields, id=None):
	if xyz.dtype != numpy.float32:
	    xyz = xyz.astype( numpy.float32 )
	#if len(xyz) == 1:
	#    xyz = xyz.reshape(1,-1)
	#    fields = fields.reshape(1,-1)
	
	n = xyz.shape[0]

	if id is None:
	    id = range(n)

	if not isinstance(id, numpy.ndarray):
	    id = numpy.array( id, dtype=numpy.int32 ).reshape( (n,) )

	if isinstance(fields, numpy.ndarray):
	    totcolumns = fields.shape[1]
	else:
	    totcolumns = 0
	    for field in fields:
		if not isinstance(field, numpy.ndarray):
		    raise ValueError("PbWriter.writepcles(): fields must be numpy arrays")
		if field.shape[0] != n:
		    raise ValueError("PbWriter.writepcles(): fields must have as many rows as xyz point array but %d != %d" % (field.shape[0], n))
		if len(field.shape) == 1:
		    totcolumns += 1
		else:
		    totcolumns += field.shape[1]

	if totcolumns != len(self.fieldnames):
	    raise ValueError("PbWriter.writepcles(): %d fieldnames but %d total columns worth of fields" % (len(self.fieldnames), totcolumns))

	pbtype = numpy.dtype( [('id', 'i4'), ('p', 'f4', (3,)), ('f', 'f4', (len(self.fieldnames), ) )] )

	pbdata = numpy.ndarray( (n,), dtype=pbtype )
	pbdata['id'] = id
	pbdata['p'] = xyz

	# arrange the fields
	if isinstance(fields, numpy.ndarray):
	    pbdata['f'] = fields.astype(numpy.float32)
	else:
	    # fields is a list/tuple of separate fields.  Dole them out
	    o = 0
	    for field in fields:
		if len(field.shape) == 1:
		    pbdata['f'][:, o] = field
		    o += 1
		else:
		    ncolumns = field.shape[1]
		    pbdata['f'][:, o:o+ncolumns] = field
		    o += ncolumns

	pbdata.tofile( self.outf )

    def close(self):
	if self.outf:
	    self.outf.close()
	self.outf = None

    def __del__(self):
	self.close()

if __name__ == "__main__":
    import numpy

    # from pbio import PbWriter

    # Make some data - four variables defined on a sphere

    thetas = numpy.linspace( 0, 2*numpy.pi, num=120, endpoint=False )
    phis = numpy.linspace( -0.5*numpy.pi, 0.5*numpy.pi, num=60, endpoint=True )

    # spherical coordinates
    npts = len(thetas) * len(phis)
    xyz = numpy.zeros( shape=(npts, 3) )		  # particle positions
    ids = numpy.zeros( shape=(npts,), dtype=numpy.int )   # integer particle ids, even though we don't need them here
    attrs = numpy.zeros( shape=(npts, 4) )		  # particle attributes

    attrnames = ['theta', 'phi', 'simplefunc', 'wildfunc']


    # There are much better ways to fill in xyz[] and attrs[], but this is simple & clear
    i = 0
    for phi in phis:
	for theta in thetas:
	    xyz[i,0] = numpy.cos(theta) * numpy.cos(phi)
	    xyz[i,1] = numpy.sin(theta) * numpy.cos(phi)
	    xyz[i,2] = numpy.sin(phi)
	    attrs[i,0] = theta
	    attrs[i,1] = phi
	    attrs[i,2] = numpy.cos( 3 * theta )	# simple function
	    attrs[i,3] = numpy.cos( 3 * theta + 3 * numpy.tan( 0.9*phi ) )  # wilder function
	    ids[i] = i
	    i += 1

    pbwriter = PbWriter( 'pbio-example.pb', attrnames )
    pbwriter.writepcles( xyz, attrs, id=ids )
    pbwriter.close()

    print "#! /usr/bin/env partiview"
    print "# Paste this into a script, e.g. pbio-example.cf, and run: partiview pbio-example.cf"
    print "pb pbio-example.pb"
    print "eval lum const 5"
    print "eval color wildfunc -1.01 1.01 # slightly bigger range than wildfunc's actual values"
    print "# \"color wildfunc\" or \"color simplefunc\" or \"color theta\" or \"color phi\""
    print "eval jump -2.5 -5.3 3.5  51 -36 -38"

